/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include <iostream>
#include <simulator.h>

using namespace std;

Simulator *simu;

const Simulator::GenericDispatcher* tick1Event, *tick2Event;

void tick1() {
    cout << "Tick1 at "  << simu->getTime() << endl;
    simu->cancel(tick1Event);
    tick1Event=simu->postCyclicIn(1.0, 1.0, &tick1);
}
void tick2() {
    cout << "Tick2 at "  << simu->getTime() << endl;
    simu->cancel(tick2Event);
    tick2Event=simu->postCyclicIn(1.0, 1.0, &tick2);
}

void maxTimeReached() {
    cout << "Max time reached at " << simu->getTime() << endl;
    simu->stop();
}

static long ctr =0;
static double tctr = 0.0;
void cyclicCounter() {
    if (tctr>simu->getTime()) {
        cerr << "Error! Sequence ordering not correct for counter=" << ctr << ", time=" << simu->getTime() << endl;
        simu->stop();
    }
    tctr = simu->getTime();
    ++ctr;
}

int main(int argc, char** argv) {

    char* ID = 0;
    if (argc>1) ID = argv[1];
    
    simu = Simulator::newInstance(ID);

    if (simu==0) {
        cout << "No simulator implementation available" << endl;
        return -1;
    }
    simu->setTimeFactor(100000.0);
    
    cout << "Validation of cancel/repost of current event" << endl;
    tick1Event = simu->postCyclic(1.0, &tick1);
    tick2Event = simu->postCyclic(1.0, &tick2);
    simu->postAt(10.0, &maxTimeReached);
    simu->start(false); // reuse main thread
    simu->cancel(tick1Event);
    simu->cancel(tick2Event);
    
    cout << "Validation of correct sequence order after MAX_EVENT" << endl;
    cout << "t=" << simu->getTime() << endl;
    simu->postCyclic(80.0/(double)Simulator::MAX_EVENTS, &cyclicCounter);
    simu->postAt(100.0, &maxTimeReached);
    simu->start(false); // reuse main thread
    cout << "t=" << simu->getTime() << ", tctr=" << tctr << ", ctr=" << ctr << endl;

    cout << "End of sequence" << endl;
}

