/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include <math.h>

#include <osg/Billboard>
#include <osg/Group>
#include <osg/Geode>
#include <osg/Geometry>
#include <osg/Texture2D>
#include <osg/TexEnv>
#include <osg/BlendFunc>
#include <osg/AlphaFunc>

#include "BillboardObstacle.h"

BillboardObstacle::BillboardObstacle( const osg::Vec3d & pos, osg::Texture2D * texture, double width, double height)
: Obstacle(pos)
{
    this->texture = texture;
    this->width = width;
    this->height = height;
    radius = sqrt(0.25*width*width + height*height); // half width and full height
}

BillboardObstacle::BillboardObstacle( const double & x, const double & y, const double & z, osg::Texture2D * texture, double width, double height )
: Obstacle(x,y,z)
{
    this->texture = texture;
    this->width = width;
    this->height = height;
    radius = sqrt(0.25*width*width + height*height); // half width and full height
}

// Code "inspired" from OSG Hangglide example
osg::Node * BillboardObstacle::buildAvatar( )
{
    osg::Vec3Array& v = *(new osg::Vec3Array(4));
    osg::Vec2Array& t = *(new osg::Vec2Array(4));
    osg::Vec4Array& l = *(new osg::Vec4Array(1));

    // color indices
    l[0].set(1,1,1,1);

    // vertices for the quad
    v[0].set(-width*0.5, 0.0, 0.0);
    v[1].set(width*0.5, 0.0, 0.0);
    v[2].set(width*0.5, 0.0, height);
    v[3].set(-width*0.5, 0.0, height);

    // texture coordinates
    t[0].set(0.0, 0.0);
    t[1].set(1.0, 0.0);
    t[2].set(1.0, 1.0);
    t[3].set(0.0, 1.0);

    osg::Geometry *geom = new osg::Geometry;

    geom->setVertexArray( &v );

    geom->setTexCoordArray( 0, &t );

    geom->setColorArray( &l );
    geom->setColorBinding( osg::Geometry::BIND_OVERALL );

    geom->addPrimitiveSet( new osg::DrawArrays(osg::PrimitiveSet::QUADS,0,4) );
    
    osg::StateSet *dstate = new osg::StateSet;
    
    dstate->setTextureAttributeAndModes(0, texture, osg::StateAttribute::ON );
    dstate->setTextureAttribute(0, new osg::TexEnv );

    dstate->setAttributeAndModes( new osg::BlendFunc, osg::StateAttribute::ON );

    osg::AlphaFunc* alphaFunc = new osg::AlphaFunc;
    alphaFunc->setFunction(osg::AlphaFunc::GEQUAL,0.05f);
    dstate->setAttributeAndModes( alphaFunc, osg::StateAttribute::ON );

    dstate->setMode( GL_LIGHTING, osg::StateAttribute::OFF );
    
    dstate->setRenderingHint( osg::StateSet::TRANSPARENT_BIN );

    geom->setStateSet( dstate );

    osg::Billboard *bb = new osg::Billboard;
    bb->addDrawable(geom);
    return bb;
}

