/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "Obstacle.h"

#include <osg/ShapeDrawable>
#include <osg/Geode>
#include "Environment.h"
using namespace Environment;

Obstacle::Obstacle( const osg::Vec3d & pos )
{
    addToLocalityDatabase(); // has no effect if not in Bricks mode
    setPosition(pos);
    transform = new TranslateTransform(pos);
    radius = 1.0;
}

Obstacle::Obstacle( const double & x, const double & y, const double & z )
{
    addToLocalityDatabase(); // has no effect if not in Bricks mode
    setPosition(x,y,z);
    transform = new TranslateTransform(x,y,z);
    radius = 1.0;
}

// modification to the scene graph is done at the right time
static void addAvatarHelper(osg::Group * group, osg::Node* transform) {
    group->addChild(transform);
}

void Obstacle::addAvatarToGroup( osg::Group * group, bool delayToNextFrame ) {
    osg::Node *avatar = buildAvatar();
    transform->addChild(avatar);
    if (delayToNextFrame) viewer->post(&addAvatarHelper, group, static_cast<osg::Node*>(transform));
    else addAvatarHelper(group, transform);
}

// Shared geometry between all the default obstacles, thanks to osg ref counting
// private to this compilation unit
static osg::Node *commonObstacleAvatar = 0;

// see comments in Agentcore
osg::Node * Obstacle::buildAvatar( )
{
    // will be ref counted when added to another parent!
    if (commonObstacleAvatar!=0) return commonObstacleAvatar;

    // create a container that makes the sphere drawable
    osg::ShapeDrawable *sd = new osg::ShapeDrawable(new osg::Sphere(osg::Vec3(0,0,0),1));
    
    sd->setColor(osg::Vec4(1.0f,1.0f,1.0f,1.0f));

    // create a geode object as a container for our drawable object
    osg::Geode* geode = new osg::Geode();
    
    // add our drawable to the geode container
    geode->addDrawable(sd);

    commonObstacleAvatar = geode;
    return commonObstacleAvatar;
}

void Obstacle::updateNeighbors( double radius )
{
    findNeighbors(radius);
}


