/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#ifndef SHORTESTANGLECORRECTION_H
#define SHORTESTANGLECORRECTION_H

#include "AttitudeCorrection.h"

/**

Uses the shortest angle rotation to map the old forward vector to the new forward vector.

This should be equivalent to minimizing the change in the up vector
in the OpenSteer Library, as we combine both of their rotations,
but this is implemented with quaternions.

No costly acos or sin are used, the quaternion is found only
with fast geometrical properties.
See for example http://logiciels.cnes.fr/MARMOTTES/marmottes-mathematique.pdf
for an explanation of the math used.
    
Care is taken for reconditionning the quaternion
after too many rotations, when this becomes necessary.

@author Nicolas Brodu
*/
class ShortestAngleAttitudeCorrection : public AttitudeCorrection
{
public:
    ShortestAngleAttitudeCorrection();

    ~ShortestAngleAttitudeCorrection();

    virtual void updateAttitude(osg::Quat &newAttitude, const osg::Quat &oldAttitude, const osg::Vec3d &newForward, const osg::Vec3d &oldForward);
};

#endif
