/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/
#include "TerrainAgent.h"
#include "KeepUpAttitudeCorrection.h"

// class private to this compilation unit
struct PlaneTerrain : public TerrainInterface {
    
    virtual double getFloorZ(const double x, const double y) {
        return 0;
    }
    
    virtual osg::Vec3d getFloorNormal(const double x, const double y) {
        return osg::Vec3d(0,0,1);
    }

};

// static to this compilaton unit
static PlaneTerrain planeTerrain;

// Default value for agents is infinite Z=0 plane floor.
TerrainInterface* TerrainAgent::terrain = &planeTerrain;

TerrainAgent::TerrainAgent()
{
    attitudeCorrection = new KeepUpAttitudeCorrection(osg::Vec3d(0,0,1), osg::Vec3d(0,0,1));
}


TerrainAgent::~TerrainAgent()
{
    delete attitudeCorrection;
}

void TerrainAgent::updateIntegration( const double elapsedTime, const double currentTime )
{
    double floorZ = terrain->getFloorZ(newPosition.x(), newPosition.y());
    floorNormal = terrain->getFloorNormal(newPosition.x(), newPosition.y());
    altitude = newPosition.z() - floorZ;
    
    // update up vector
    ((KeepUpAttitudeCorrection*)attitudeCorrection)->targetUp = floorNormal;

    if (altitude < 0) {
        newPosition.z() = floorZ;
        ((KeepUpAttitudeCorrection*)attitudeCorrection)->matchTarget(newAttitude,newForward);
    }
}

void TerrainAgent::updateAI( const double elapsedTime, const double currentTime )
{
    double floorZ = terrain->getFloorZ(getPosition().x(), getPosition().y());
    floorNormal = terrain->getFloorNormal(getPosition().x(), getPosition().y());
    altitude = getPosition().z() - floorZ;
        
    // update up vector
    ((KeepUpAttitudeCorrection*)attitudeCorrection)->targetUp = floorNormal;
}


