/***************************************************************************
 *   Copyright (C) 2004 by Nicolas Brodu                                   *
 *   nicolas.brodu@free.fr                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/


#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <osg/Geometry>
#include <osg/ShapeDrawable>
#include <osgProducer/Viewer>
#include <osgGA/GUIEventHandler>
#include <osgDB/ReadFile>

#include <osg/Material>
#include <osg/Geode>
#include <osg/BlendFunc>
#include <osg/Depth>
#include <osg/Projection>
#include <osg/PolygonOffset>
#include <osg/MatrixTransform>

#include <osgText/Text>


#include <iostream>
#include <sstream>

#include <time.h>

#include "Environment.h"
#include "BillboardObstacle.h"
#include "Terrain.h"
#include "CrogaiAgent.h"

using namespace std;
using namespace Environment;

Terrain* theFloor = 0;

osgText::Text* scoresText = 0;
osgText::Text* energyText = 0;

// Directly copied from the OSG examples
osg::Node* createHUD()
{
    osg::Geode* geode = new osg::Geode();
    
    // turn lighting off for the text and disable depth test to ensure its always ontop.
    osg::StateSet* stateset = geode->getOrCreateStateSet();
    stateset->setMode(GL_LIGHTING,osg::StateAttribute::OFF);

    // Disable depth test, and make sure that the hud is drawn after everything
    // else so that it always appears ontop.
    stateset->setMode(GL_DEPTH_TEST,osg::StateAttribute::OFF);
    stateset->setRenderBinDetails(11,"RenderBin");

    scoresText = new osgText::Text;
    geode->addDrawable( scoresText );
    scoresText->setFont("FreeSans.ttf");
    scoresText->setPosition(osg::Vec3(WindowWidth/1000.0,WindowHeight-WindowHeight/100.0,0.0f));
    scoresText->setText("");
    //scoresText->setCharacterSizeMode(osgText::SCREEN_COORDS);
    scoresText->setCharacterSize(WindowHeight/25.0);
    scoresText->setAlignment(osgText::Text::LEFT_TOP);

    energyText = new osgText::Text;
    geode->addDrawable( energyText );
    energyText->setFont("FreeSans.ttf");
    energyText->setPosition(osg::Vec3(WindowWidth - WindowWidth/1000.0,WindowHeight-WindowHeight/100.0,0.0f));
    energyText->setText("");
    //scoresText->setCharacterSizeMode(osgText::SCREEN_COORDS);
    energyText->setCharacterSize(WindowHeight/25.0);
    energyText->setAlignment(osgText::Text::RIGHT_TOP);

    osg::BoundingBox bb(osg::Vec3(0,WindowHeight-WindowHeight/25.0-WindowHeight/100.0,0),osg::Vec3(WindowWidth,WindowHeight,0));
    osg::Geometry* geom = new osg::Geometry;

    osg::Vec3Array* vertices = new osg::Vec3Array;
    float depth = bb.zMin()-0.1;
    vertices->push_back(osg::Vec3(bb.xMin(),bb.yMax(),depth));
    vertices->push_back(osg::Vec3(bb.xMin(),bb.yMin(),depth));
    vertices->push_back(osg::Vec3(bb.xMax(),bb.yMin(),depth));
    vertices->push_back(osg::Vec3(bb.xMax(),bb.yMax(),depth));
    geom->setVertexArray(vertices);

    osg::Vec3Array* normals = new osg::Vec3Array;
    normals->push_back(osg::Vec3(0.0f,0.0f,1.0f));
    geom->setNormalArray(normals);
    geom->setNormalBinding(osg::Geometry::BIND_OVERALL);

    osg::Vec4Array* colors = new osg::Vec4Array;
    colors->push_back(osg::Vec4(1.0f,1.0,0.8f,0.2f));
    geom->setColorArray(colors);
    geom->setColorBinding(osg::Geometry::BIND_OVERALL);

    geom->addPrimitiveSet(new osg::DrawArrays(GL_QUADS,0,4));
    
    osg::StateSet* stateset2 = geom->getOrCreateStateSet();
    stateset2->setMode(GL_BLEND,osg::StateAttribute::ON);
    //stateset2->setAttribute(new osg::PolygonOffset(1.0f,1.0f),osg::StateAttribute::ON);
    stateset2->setRenderingHint(osg::StateSet::TRANSPARENT_BIN);

    geode->addDrawable(geom);

    // create the hud.
    osg::MatrixTransform* modelview_abs = new osg::MatrixTransform;
    modelview_abs->setReferenceFrame(osg::Transform::ABSOLUTE_RF);
    modelview_abs->setMatrix(osg::Matrix::identity());
    modelview_abs->addChild(geode);

    osg::Projection* projection = new osg::Projection;
    projection->setMatrix(osg::Matrix::ortho2D(0,WindowWidth,0,WindowHeight));
    projection->addChild(modelview_abs);

    return projection;

}


struct PredatorAgent : public CrogaiAgent {

    PredatorAgent(int specie): CrogaiAgent(specie) {}

    void updateAI( const double elapsedTime, const double currentTime) {
        CrogaiAgent::updateAI(elapsedTime,currentTime);
        
        // avoidance force. Avoid only the closest object
        // Since neighbors are returned sorted, this is easy
        osg::Vec3d avoidForce(0,0,0);

        // Forces for Boid-like behavior
        osg::Vec3d localCenter(0,0,0);
        osg::Vec3d forwardForce(0,0,0);

        // Predator/prey force
        osg::Vec3d seekFleeForce(0,0,0);
        
        int nSeekFlee = 0; // number of neighbor preys + predators
        int n = 0; // number of neighbor agents
        
        findNeighbors(searchRadius);
        neighborsAccessLock();
        for (multimap<const double,Localized*>::iterator it = neighbors->begin(); it != neighbors->end(); ++it) {

                if (wrappedDiffVector(it->second->getPosition(), getPosition()) * forward < 0)
                    continue; // see only what's in front of us

                Agent* neighbor = dynamic_cast<Agent*>(it->second);

                // if not an agent, maybe it is an obstacle?
                if (!neighbor) {
                    Obstacle* obs = dynamic_cast<Obstacle*>(it->second);
                    // Avoid obstacles, but only the closest one for which there is a collision risk
                    if (obs!=0) {
                        avoidForce = steerToAvoid(obs->getNode(), searchRadius);
                        break;
                    }
                    continue;
                }


                PredatorAgent* sa = dynamic_cast<PredatorAgent*>(neighbor);
                // Escape from unknown things from outer space!!!
                if (sa==0) {
                    steeringForce += steerForEvasion(neighbor);
                }
                // One of our mates. Boidizing...
                else if (sa->specie == specie) {
                    // Check for risk of collision
                    avoidForce = steerToAvoid(neighbor, searchRadius);
                    
                    // Collision risk => no need to look for further AI
                    if ((avoidForce.x()!=0) || (avoidForce.y()!=0) || (avoidForce.z()!=0))
                        break;
    
                    steeringForce += steerForFlee(neighbor->getPosition()) / (it->first+0.001);
                    forwardForce += neighbor->forward;
                    localCenter += neighbor->getPosition();
                    ++n;
                }
                // Hey, another guy! Let's try to catch it unaware...
                else {
                    // hehe, we are behind, it can't see us
                    if (sa->forward * forward < 0) {
                        seekFleeForce += steerForPursuit(sa);
                        ++nSeekFlee; // flag
                        break;
                    }
                    // Oops, avoid fight
                    else {
                        seekFleeForce += steerForEvasion(sa);
                        ++nSeekFlee; // flag
                        break;
                    }
                }
        }
        neighborsAccessUnlock();
        
        // Collision risk? Priority over anything else
        if ((avoidForce.x()!=0) || (avoidForce.y()!=0) || (avoidForce.z()!=0)) {

            // Don't try to avoid by digging a hole, it doesn't work!
            // When not flying, project avoidance direction on the horizontal plane
            if (altitude<radius) {
                steeringForce = avoidForce - floorNormal * (avoidForce*floorNormal);
                steeringForce *= sqrt(getMaxForceSquared() / steeringForce.length2());
            }
            else steeringForce = avoidForce;

            return;
        }

        // don't move to the origin when no other agent present
        // In this case, move at random, but only if not already doing something else
        if ((n==0) && (nSeekFlee==0)) steeringForce += steerForWander(0.4);

        // Otherwise, agent behavior when in company of friends
        // - steer to avoid neighbors
        // - steer toward the local center
        // - and also to keep the local group direction
        else if (n>0) {
            steeringForce /= (double)n; // average of avoidance vectors
            steeringForce += steerForSeek(localCenter/(double)n) * 2; // local center
            steeringForce += forwardForce /(double)n  -  forward; // local group direction
        }

        // prey/predator avoidance?
        if (nSeekFlee>0) {
            steeringForce += seekFleeForce * 10; // give it more weight than boidizing
        }
        
        // If flying, try to avoid floor, up to max height were we're "safe"
        if (altitude>radius) {
            if ( altitude < radius*5.0) 
                steeringForce += floorNormal * getMaxForce();
        }
        
        // Otherwise, have a 1% chance to take off at random, supposing gravity doesn't bring us back
        else {
            if (Utility::Random() < 0.01) steeringForce += floorNormal * getMaxForce();
        }
        
    }


    virtual void winnerAction(CrogaiAgent* looser) {
        CrogaiAgent::winnerAction(looser);
        if (scoresText==0) return;

        ostringstream os;

        os << "Pop.: ";
        allAgentsAccessLock();
        os << allAgents.size();
        allAgentsAccessUnlock();
        os << ". Scores: ";
        for (map<int, int>::iterator it = scores.begin(); it != scores.end(); ++it) {
            if (it!=scores.begin()) os << ", ";
            os << it->second;
        }
        scoresText->setText(os.str());
        
        ostringstream os2;
        os2 << "Winner energy: " << energy;
        energyText->setText(os2.str());
        
    }
    
};

osg::Group* root = 0, *metaroot = 0;
osg::Group *normalGroup = 0;

class KeyboardEventHandler : public osgGA::GUIEventHandler {
public:

    void translateRoot(int tx, int ty) {
        double s = theFloor->getWorldSize();
        TranslateTransform* tt=0;
        tt = new TranslateTransform(tx*s, ty*s, 0);
        tt->addChild(root);
        metaroot->addChild(tt);
    }

    virtual bool handle(const osgGA::GUIEventAdapter& ea, osgGA::GUIActionAdapter&) {
        switch(ea.getEventType()) {

            case osgGA::GUIEventAdapter::KEYDOWN:
                switch(ea.getKey()) {
                    case 'g':
                        gravity += 0.1;
                        cout << "gravity is now " << gravity << endl;
                        break;
                    case 'G':
                        gravity -= 0.1;
                        if (gravity<0) gravity=0;
                        cout << "gravity is now " << gravity << endl;
                        break;
                    
                    case 'p':
                        simu->stop();
                        break;
                    case 'P':
                        simu->start();
                        break;

                    case 'n':
                    case 'N':
                        if (normalGroup==0) {
                            normalGroup = theFloor->drawNormals();
                            root->addChild(normalGroup);
                            cout << "Floor normals on" << endl;
                        }
                        else {
                            root->removeChild(normalGroup);
                            normalGroup = 0;
                            cout << "Floor normals off" << endl;
                        }
                        break;

                    case 'C':
                    case 'c':
                        // keep only the first child, root
                        if (metaroot->getNumChildren() > 9) {
                            metaroot->removeChild(1, metaroot->getNumChildren() -1);
                            cout << "Cyclic terrain visualization off" << endl;
                            break;
                        }
                        if (metaroot->getNumChildren() >1 ) {
                            translateRoot(-2,-2);
                            translateRoot(-2,-1);
                            translateRoot(-2,0);
                            translateRoot(-2,1);
                            translateRoot(-2,2);
                            translateRoot(2,-2);
                            translateRoot(2,-1);
                            translateRoot(2,0);
                            translateRoot(2,1);
                            translateRoot(2,2);
                            translateRoot(-1,-2);
                            translateRoot(0,-2);
                            translateRoot(1,-2);
                            translateRoot(-1,2);
                            translateRoot(0,2);
                            translateRoot(1,2);
                            cout << "Cyclic terrain visualization: level 2" << endl;
                            break;
                        }
                        translateRoot(1,0);
                        translateRoot(-1,0);
                        translateRoot(0,1);
                        translateRoot(0,-1);
                        translateRoot(1,1);
                        translateRoot(-1,1);
                        translateRoot(1,-1);
                        translateRoot(-1,-1);

                        cout << "Cyclic terrain visualization: level 1" << endl;
                        break;

                    
                    default:
                        return false;
                }
                return true;
/*
            case osgGA::GUIEventAdapter::KEYUP:
                return true;
*/
            default:
                return false;
        }
    }

    virtual void accept(osgGA::GUIEventHandlerVisitor& v) {
            v.visit(*this);
    }

};

int main(int argc, char *argv[])
{
    osg::Referenced::setThreadSafeReferenceCounting(true);

    simu = Simulator::newInstance(); //new OpenThreadSimulator;// POSIX_Simulator();

    createViewer("Predator match");

    root = new osg::Group();
    
//    theFloor = new Terrain(800, 100, 7, 0.3, 3.0);
//    theFloor = new Terrain(400, 50, 7, 0.3, 3.0);

    theFloor = new Terrain(500, 100, 7, 0.35, 3.5);
    TerrainAgent::terrain = theFloor;
    root->addChild(theFloor);
    

    //theFloor->debugNormals();

    // create new keyboard event handler
    KeyboardEventHandler* keh = new KeyboardEventHandler();
    viewer->getEventHandlerList().push_front(keh);

    metaroot = new osg::Group();
    metaroot->addChild(root);

    osg::Group * wholeScene = new osg::Group();
    wholeScene->addChild(metaroot);
    wholeScene->addChild(createHUD());

    viewer->setSceneData( wholeScene );
    viewer->realize();
    
    //Localized::setDefaultLocalityMode(Localized::Mixed);
    Localized::setDefaultLocalityMode(Localized::Bricks);
    Localized::createLocalityDatabase(
        -theFloor->getHalfWorldSize(), -theFloor->getHalfWorldSize(), 0,
        theFloor->getWorldSize(), theFloor->getWorldSize(), theFloor->getMaxHeight()*2.0,
        100,100,50);

    AgentCore::setDefaultMaxForce(20);
    AgentCore::setDefaultMaxSpeed(8);
    //AgentCore::setDefaultMaxEnergy(200);
    gravity=16;

    double istep = 0.04; // 25 f/s
    double aistep = .15;
    simu->setTimeFactor(1.0);
    AgentCore::setDefaultEnergyFootprintRatio(0.001);
    AgentCore::setDefaultInitialEnergy(500.0);

// Obstacle code
    const int nobstacles = 100;

    osg::Texture2D *texture0 = new osg::Texture2D;
    texture0->setImage(osgDB::readImageFile("tree.rgba"));
    osg::Texture2D *texture1 = new osg::Texture2D;
    texture1->setImage(osgDB::readImageFile("buisson.tga"));
    for (int i=0; i<nobstacles; ++i) {
        double x,y,z;
        do {
            x = theFloor->getWorldSize()*Utility::Random() - theFloor->getHalfWorldSize();
            y = theFloor->getWorldSize()*Utility::Random() - theFloor->getHalfWorldSize();
            z = theFloor->getFloorZ(x,y);
        }while (z * Utility::Random() > theFloor->getMaxHeight() * 0.25);

        Obstacle *obs;
        switch (Utility::IRandom(0,5)) {
            case 0:
            case 1:
            case 2:
                obs = new BillboardObstacle(x, y, z, texture0,
                    2.0 + 2.0*Utility::Random(),
                    4.0 + 2.0*Utility::Random()
                );
                break;
            case 3:
            case 4:
                obs = new BillboardObstacle(x, y, z, texture1,
                    2.0 + Utility::Random(),
                    2.0 + Utility::Random()
                );
                break;
            case 5:
                obs = new Obstacle(x, y, z);
                break;
        };
        // Position Obstacle at floor surface
        obs->addAvatarToGroup(root);

        // Obstacles update their neighbor lists to help agents, in mixed mode only
        /*
        simu->postCyclicAt<Obstacle>(
            // add jitter so all ai computations not done at the same time
            aistep * Utility::Random(),  
            aistep, obs, &Obstacle::updateNeighbors, 15.0);
        */
    }

    
// Agent code
    const int nagents = 600;
    const int nspecies = 5;
    for (int i=0; i<nagents; ++i) {
        PredatorAgent *agent = new PredatorAgent((i%nspecies) + 1);
        double x = theFloor->getWorldSize()*Utility::Random() - theFloor->getHalfWorldSize();
        double y = theFloor->getWorldSize()*Utility::Random() - theFloor->getHalfWorldSize();
        // Position Agent at floor surface
        agent->setPosition(x,y, theFloor->getFloorZ(x,y));
        agent->addAvatarToGroup(root);

        agent->aiEvent = simu->postCyclicAt<Agent>(
            // add jitter so all ai computations not done at the same time
            aistep * Utility::Random(),
            aistep, agent, &Agent::aiStep);
        
        agent->integrationEvent = simu->postCyclicAt<Agent>(
            // add jitter so all integration step computations not done at the same time
            istep * Utility::Random(),
            istep, agent, &Agent::integrationStep);
        
        // add roughly one fourth of the agents to the locality database
        // the others keep their own neighborhood
        //if (Utility::Random() < 0.5)
            agent->addToLocalityDatabase();
        
    }
    
    // Provide an initial list of neighbors to all agents
    for (list<Localized*>::iterator it = Localized::allObjects.begin(); it != Localized::allObjects.end(); ++it) {
        (*it)->findNeighbors(30.0);
    }
    // Second order search
    for (list<Localized*>::iterator it = Localized::allObjects.begin(); it != Localized::allObjects.end(); ++it) {
        (*it)->findNeighbors(30.0);
    }

    simu->start(); // in its own thread
    
    while( !viewer->done() )
    {
        viewer->sync();
        viewer->update();
        viewer->frame();
    }
    // wait for all OSG threads to sync
    viewer->sync();
    
    // flush all pending simu events (cyclic...)
    simu->flush();
    
    // wait for simu to stop
    simu->stop(true);
    
    exit(0); // bug in osg/openthreads at object deletion. Use exit instead of return
    return 0;
}



