#!/usr/bin/perl

sub help() {
    print "Missing or bad arguments: argument: cycle_min_amplitude pop_log1 [pop_log2 ...]\n";
    print "\n";
    print "This script should be run on the population log files. If you have no better idea, try using the initial population count for a specie as cycle_min_amplitude, 80 for this version of the programs.\n";
    exit(1);
}


$min_amplitude = $ARGV[0];
##$min_half_period = 0.5 * $ARGV[1];


if ( $min_amplitude == 0 ) {
    help();
}

## Experimental time-frequency analysis => uncomment all ## comments to use
##if ( $min_half_period == 0 ) {
##    help();
##}


if ($#ARGV < 2) {
    help();
}

# remove the cycle argument
shift @ARGV; ##shift @ARGV;

$meanAveragePerSpecie = 0;
$varAveragePerSpecie = 0;

# loop over file names
foreach $fileName (@ARGV) {
    
    open(FILE,"<$fileName") or die "Cannot open $fileName for reading.\n";

@names = ();
@minValues = ();
@maxValues = ();
@minTimes = ();
@maxTimes = ();
@ncycles = ();
@phase = ();


LOOP: while (<FILE>) {

    chomp;

    /^#.*/ && next LOOP;
    /^$/ && next LOOP;

    # init
    if (@names==()) {
        @names = split;
        # force allocation of 0 for all species
        for ($i=0; $i<$#names; ++$i) {$ncycles[$i] = 0; $phase[$i] = 0;}
        next LOOP;
    }

    if (@minValues==()) {
        @minValues = split;
        $t = shift @minValues; # discard time
        @maxValues = split;
        shift @maxValues; # discard time
        for ($i=0; $i<=$#minValues; ++$i) {$minTimes[$i] = $t; $maxTimes[$i] = $t;}
        next LOOP;
    }

    @entry = split;

    # treat time separatly
    for ($i=0; $i<$#entry; ++$i) {
        if ($entry[$i+1] < $minValues[$i]) {
            $minValues[$i] = $entry[$i+1];
            $minTimes[$i] = $entry[0];
            if ($maxValues[$i] - $minValues[$i] > $min_amplitude) {$phase[$i] = 1;}
        }
        if ($entry[$i+1] > $maxValues[$i]) {
            $maxValues[$i] = $entry[$i+1];
            $maxTimes[$i] = $entry[0];
            if ($maxValues[$i] - $minValues[$i] > $min_amplitude) {$phase[$i] = 1;}
        }
        
        #if ($maxValues[$i] - $minValues[$i] > $min_amplitude) {$phase[$i] = 1;}
        
        # note when going back below threshold and cycle condition met => cycle
        if ($phase[$i] == 1) {
            # coming back from a maximum
            if (($maxTimes[$i] > $minTimes[$i]) && ($maxValues[$i] - $entry[$i+1] >  $min_amplitude)) ## && ($maxTimes[$i] - $minTimes[$i] > $min_half_period))
            # and period is enough
            # then cycle
            {
                $phase[$i] = 0;
                $ncycles[$i]++;
                # reset min for next local optima
                $minValues[$i] = $entry[$i+1];
                $minTimes[$i] = $entry[0];
            }
            # or a minumum too
            elsif (($maxTimes[$i] < $minTimes[$i]) && ($entry[$i+1] - $minValues[$i] > $min_amplitude)) ## && ($minTimes[$i] - $maxTimes[$i] > $min_half_period))
            # and period is enough
            # then cycle
            {
                $phase[$i] = 0;
                # don't count cycle twice, count bumps and not valleys
                # $ncycles[$i]++;
                # reset max for next local optima
                $maxValues[$i] = $entry[$i+1];
                $maxTimes[$i] = $entry[0];
            }
        }
    }

    

}
close(FILE);

print "# Cycles per specie: ";
print join(" ", @ncycles)."\n";
$average = 0; map { $average += $_} @ncycles;
$average /= 1 + $#ncycles;
print "# Average per specie: $average\n";

$meanAveragePerSpecie += $average;
$varAveragePerSpecie += $average * $average;

} # end of file loop

# finish computations
$numFiles = 1 + $#ARGV;
$meanAveragePerSpecie /= $numFiles;
$varAveragePerSpecie = sqrt( ($varAveragePerSpecie / $numFiles) - $meanAveragePerSpecie * $meanAveragePerSpecie );

print "Mean number of cycles per specie: $meanAveragePerSpecie, standard deviation: $varAveragePerSpecie \n";
