#!/usr/bin/perl

sub help() {
    print "Missing or bad arguments: argument: statistics_XXX.txt [statistics_YYY.txt ...]\n";
    print "\n";
    print "This script should be run on the statistics log files. It uses the corresponding populations file (same random seed) to check on which species went extinct, and monitor how the initial agents turnover ratio for species than went extinct.\n";
    exit(1);
};

if ($#ARGV < 0) {
    help();
}

# ratios for (mean over std dev) and (mean over time before extinction)
$meanRatioVM = 0;
$varRatioVM = 0;
$meanRatioMT = 0;
$varRatioMT = 0;

$totalExtinctionNumber = 0;

# loop over file names
foreach $fileName (@ARGV) {

    if (!($fileName =~ /.*statistics.*/)) {
        print "Incorrect file name: $fileName\n";
        help();
    }
    
    open(FILE,"<$fileName") or die "Cannot open $fileName for reading.\n";
    
    $popFile = $fileName;
    $popFile =~ s/statistics/populations/g;
    open(POPFILE,"<$popFile") or die "Cannot open $popFile for reading.\n";

$first = 0;
@popEntry = ();
@popTimes = ();
POPLOOP: while (<POPFILE>) {

    chomp;

    /^#.*/ && next POPLOOP;
    /^$/ && next POPLOOP;

    # init: skip names
    if ($first==0) {
        $first = 1;
        @names = split;

        # force allocation of 0 max time for all species
        for ($i=0; $i<$#names; ++$i) {$popTimes[$i] = 0;}
        next POPLOOP;
    }
    
    #discard time => check only on extinctions. Use it here
    @popEntry = split;
    $time = shift @popEntry;
    
    for ($i=0; $i<=$#popEntry; ++$i) {
        if ($popEntry[$i] != 0) {$popTimes[$i] = $time;}
    }
}
close(POPFILE);

# force allocation of 0 for all species
@meanLifeTimeSpecie = ();
@stdDevLifeTimeSpecie = ();
for ($i=0; $i<=$#popEntry; ++$i) {$meanLifeTimeSpecie[$i] = 0; $stdDevLifeTimeSpecie[$i] = 0;}

$first = 0;
LOOP: while (<FILE>) {

    chomp;

    /^#.*/ && next LOOP;
    /^$/ && next LOOP;

    # init: skip names
    if ($first==0) {
        $first = 1;
        next LOOP;
    }

    @entry = split;
    
    # discard generation
    shift @entry;
    
    for ($i=0; $i<=$#popEntry; ++$i) {
        $meanLifeTimeSpecie[$i] = $entry[$i * 8];
        $stdDevLifeTimeSpecie[$i] = $entry[1 + $i * 8];
    }
    
    # only process initial generation
    last LOOP;
}
close(FILE);

print "# Life time ratios (Mean/StdDev, Mean/Time):";
for ($i=0; $i<=$#popEntry; ++$i) {
    # only monitor extinct species, final explosion for surviving specie has too much influence
    # let's compare what's comparable!
    if ($popEntry[$i]==0) {
        $ratio = $stdDevLifeTimeSpecie[$i] / $meanLifeTimeSpecie[$i];
        print " ($ratio, ";
        $meanRatioVM += $ratio;
        $varRatioVM += $ratio * $ratio;

        $ratio = $meanLifeTimeSpecie[$i] / $popTimes[$i];
        print "$ratio)";
        $meanRatioMT += $ratio;
        $varRatioMT += $ratio * $ratio;
        
        $totalExtinctionNumber ++;
    } else {
        print " (still alive)";
    }
}
print "\n";

} # end of file loop

# finish computations
$meanRatioVM /= $totalExtinctionNumber;
$varRatioVM = sqrt( ($varRatioVM / $totalExtinctionNumber) - $meanRatioVM * $meanRatioVM);
$meanRatioMT /= $totalExtinctionNumber;
$varRatioMT = sqrt( ($varRatioMT / $totalExtinctionNumber) - $meanRatioMT * $meanRatioMT);

print "Mean (std. dev. life time / average) ratio: $meanRatioVM, std. dev: $varRatioVM\n";
print "Mean (average life time / extinction time) ratio: $meanRatioMT, std. dev: $varRatioMT\n";
